"""
The sci package contains helper and wrapper functions
for scientists  
"""
__version__ = "0.1.3"

#from sci.store import *
#from sci.db import *
#from sci.compute import *
#from sci.json import *
#from sci.new import *
#from sci.old import *
#from sci.tools import *

import sci.compute
import sci.store
import sci.db
import sci.fh
import sci.json
import sci.tools
import sci.old
import sci.new
